/* 
 * Controller for 8x8 LED Matrix
 * 
 * V1.0  
 * 
 * By: Lars Ivar Miljeteig, December 2009
 * Platform: AVR ATmega8 (avr-gcc)
 * Dependencies: larslib.h, larslib.c
 *
 * Written for a 16-pin 8x8 LED Matrix (788ASR)
 *  using multiplexing. 
 * 
 * Read more at www.larsivar.com/cp/
 */


#include "mat88.h"
#include "mat88_abc.h"
#define M8_INACTIVE    0
#define M8_STATIC      1
#define M8_BLINKWRITE  2
#define M8_SCROLLWRITE 3

void (*mat88_actions[4])();

// Struct for scrolling function
struct
{
    uint8_t speed, rounds, textpos, textlen, shifted, round;
    uint16_t timer;
    uint8_t *newletter, *oldletter;
    uint8_t text[50];
} swobj;

// Struct for blinking function  
struct
{
    uint8_t speed, rounds, textpos, textlen, round;
    uint16_t timer;
    uint8_t text[50];
} blobj;

uint8_t arrows_up[] = { 
    
    0b00011000,
    0b00100100,
    0b01000010,
    0b10000001,
    0b00011000,
    0b00100100,
    0b01000010,
    0b10000001,
};


uint8_t row=0, state=0, itstime = 0;
volatile uint8_t *matrix;
uint16_t timer = 0;
uint8_t speed = 200, toggle = 0xf0;

uint8_t K[8] = {0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01};
uint8_t A[8] = {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};

/*
 * Timer interrupt for PWM multiplexing
 * of LEDs in matrix
 *
 */
ISR(TIMER0_COMPA_vect)
{
    uint8_t i;

    timer++;
    if(timer>255-speed) // TODO: Fix variable speeds!
    {
        itstime = 1; 
        timer = 0;
    }
    
    toggle = ~toggle;

    if(state != M8_INACTIVE)
    {
        if(toggle == 0xf0)
        {
            // Convert matrix into byte-map 
            for(i=0; i<8; i++) 
            {
                K[i] = 1;
                A[i] = (matrix[row]>>i) & 0x01; // flip vertical here
            }
            K[7-row] = 0; // flip horizontal here

            // Output everything:
            PORTB = (A[4]<<PB4) | (A[6]<<PB3) | (A[7]<<PB0) | (K[0]<<PB5) | (K[5]<<PB1) | (K[6]<<PB2);
            PORTC = (A[1]<<PC1) | (A[3]<<PC3) | (K[1]<<PC0) | (K[7]<<PC2);
            PORTD = (A[0]<<PD4) | (A[2]<<PD5) | (A[5]<<PD6) | (K[2]<<PD2) | (K[3]<<PD7) | (K[4]<<PD3);

            DDRB = 0x3f;
            DDRD = 0xfc;

            if(++row>7) row = 0;
        }
        else
        {
            PORTC = (A[1]<<PC1) | (A[3]<<PC3) | (K[1]<<PC0) | (K[7]<<PC2);
            DDRD = 0x00;
            DDRB = 0x00;
        }
    }
    TCNT0 = 0;
    
}

// Low-level ISR functions for state machine:

// Do nothing if current state is inactive
void mat88_ISRinactive()
{
    return;
}

// Do nothing if current state is static
void mat88_ISRstatic()
{
    return;
}

/*
 *  Scroll write, called by interrupt
 */
void mat88_ISRscrollwrite()
{
    swobj.timer = 0;
    uint8_t i = 0;

    // Make an output matrix based on the current letter and the upcomming one
    for(i=0; i<8; i++) 
    {
        matrix[i]= swobj.oldletter[i]<<(swobj.shifted) | (swobj.newletter[i]>>(8-swobj.shifted));
    }

    // If the current letter is shifted at max, insert the new letter in 
    // the current letter container. This creates a handy loop. 
    if(++swobj.shifted >= 8)
    {
        swobj.shifted = 0;
        swobj.oldletter = swobj.newletter;

        if(++swobj.textpos<swobj.textlen)
        {
            swobj.newletter = &abc[8*(swobj.text[swobj.textpos])];
        }
        else if(++swobj.round<swobj.rounds)
        {
            swobj.textpos = 0;
            swobj.newletter = &abc[8*(swobj.text[swobj.textpos])];
        }
        else
        {
            // Finished!
            state = M8_INACTIVE;
        }
    }
}

/*
 *  Blink write, called by interrupt
 */
void mat88_ISRblinkwrite()
{

    if(blobj.textpos<blobj.textlen)
    {
        matrix = &abc[8*blobj.text[blobj.textpos]];
    }
    else if(++blobj.round<blobj.rounds)
    {
        blobj.textpos = 0;
        matrix = &abc[8*blobj.text[blobj.textpos]];
    }
    else
    {
        // Finished!
        state = M8_INACTIVE;
    }
    blobj.textpos++;

}

/*
 *  Initialising
 *  Must be called for the matrix to work
 */
void mat88_init()
{
    // Set all ports as output
    PORTB = 0xff; 
    PORTC = 0xff;
    PORTD = 0xff;
    DDRB = 0x3f;
    DDRC = 0x0f;
    DDRD = 0xfc;

    TCCR0A = 0;

    // Use normal timer mode, clock/64, no PWM or fancy stuff enabled
    TCCR0B = (1<<CS01)|(1<<CS00);

	  // Enable timer compare A interrupt
	  TIMSK0 |= (1<<OCIE0A);

    // Run compare ISR double speed of overflow
    OCR0A = 127;

    // Link state machine to appropriate function pointers
    mat88_actions[M8_INACTIVE] = &mat88_ISRinactive;
    mat88_actions[M8_STATIC] = &mat88_ISRstatic;
    mat88_actions[M8_BLINKWRITE] = &mat88_ISRblinkwrite;
    mat88_actions[M8_SCROLLWRITE] = &mat88_ISRscrollwrite;
}

// High-level functions

/*
 * Draw
 * Used to output a custom caracter to the matrix
 *
 * inmat - A pointer to an array containing 8 bytes to output, 
 *         one byte per line on the matrix
 */
void mat88_draw(uint8_t *inmat)
{
    matrix = inmat;
    state = M8_STATIC;  // Activate ISR state
    return;
}

/*
 * Scroll write
 * Called to activate scroll mode, writing letters scrolling
 * past the LED matrix right to left
 *
 * Parameters: 
 * text - Text to display
 * scrollspeed - Speed to scroll the letters
 * rounds - Number of times to repeat, 0 for ethernal loop
 */
void mat88_scrollwrite(uint8_t *text, uint8_t scrollspeed, uint8_t rounds)
{
    uint8_t tl = 0;

    // Format text to array numbers
    while(text[tl]!=0)
    {
        if(text[tl]>0x40 && text[tl]<0x5B) swobj.text[tl] = text[tl]-0x41; // CAPS
        else if(text[tl]>0x60 && text[tl]<0x7B) swobj.text[tl] = text[tl]-0x61; // text
        //else if(text[textlen] == (uint8_t)' ') swobj.text[textlen] = 26; // Space
        else swobj.text[tl] = 26; // Space (default)
        tl++;
    }

    swobj.textlen = tl;
    swobj.speed = scrollspeed;
    swobj.rounds = rounds; 
    swobj.textpos = 1;
    swobj.timer = 0;
    swobj.shifted = 0;
    swobj.round = 0;

    swobj.oldletter = &abc[8*(swobj.text[0])];
    swobj.newletter = &abc[8*(swobj.text[1])];

    speed = scrollspeed;
    state = M8_SCROLLWRITE; // Activate ISR state
}

/*
 * Blink write
 * Called to activate blink mode, writing letters one at a time
 * with a short pause between each letter
 *
 * Parameters: 
 * text - Text to display
 * blinkspeed - Speed to blink the letters
 * rounds - Number of times to repeat, 0 for ethernal loop
 */
void mat88_blinkwrite(uint8_t *text, uint8_t blinkspeed, uint8_t rounds)
{
    uint8_t tl = 0;

    // Format text to array numbers
    while(text[tl]!=0)
    {
        if(text[tl]>0x40 && text[tl]<0x5B) blobj.text[tl] = text[tl]-0x41; // CAPS
        else if(text[tl]>0x60 && text[tl]<0x7B) blobj.text[tl] = text[tl]-0x61; // text
        //else if(text[tl] == (uint8_t)' ') blobj.text[tl] = 26; // Space
        else blobj.text[tl] = 26; // Space (default)
        tl++;
    }

    blobj.textlen = tl;
    blobj.speed = blinkspeed;
    blobj.rounds = rounds; 
    blobj.textpos = 0;
    blobj.timer = 0;
    blobj.round = 0;

    // Start by clearing matrix
    matrix = nothing;

    speed = blinkspeed;
    state = M8_BLINKWRITE; // Activate ISR state
}

/*
 *  Put a single character on the matrix
 */
void mat88_putc(uint8_t byte)
{
    // Encapsulation for the draw function using the huge abc-array
    mat88_draw(&abc[8*(byte-65)]);
}

/*
 *  Clear - Clears the contents of the matrix, turning off all LEDs
 */
void mat88_clear()
{
    matrix = nothing;
    state = M8_STATIC; // Activate ISR state
}

/*
 *  Refresh
 *  Call this in the main loop as often as possible, to refresh the actual
 *  text output. This function is timed by an timer interrupt flag, so
 *  it can be called asynchronously as often as desired
 */
void mat88_refresh()
{
    if(itstime) 
    {
        // Prepare the matrix
        mat88_actions[state]();
        itstime = 0;
    }
}
